import inspect
import json
import pickle
import sys
import os
import warnings

from flask import Flask, request, render_template, jsonify
from colorama import init
from colorama import Fore, Back, Style

from input_and_output import df_to_df, parse_json, preprocess

init()

try:
    import pandas as pd
except:
    warnings.warn("Could not import pandas")


class YhatModel(object):
    """
    A simple class structure for your model. Place all of the code you want to
    execute in the `execute` method.

    The `run` method provides you with a basic development "server" for testing
    locally. It will give you the same results as you will see on a Yhat server
    99% of the time.

    Parameters
    ----------
    filename: string
        filename of saved model
    """
    def __init__(self, filename=None):
        if filename is not None:
            pickles = json.load(open(filename, 'rb'))
            for varname, pickled_value in pickles.get('objects', {}).items():
                globals()[varname] = pickle.loads(pickled_value)

    def execution_plan(self, data):
        return self.execute(data)

    @preprocess
    def execute(self, data):
        """
        This is the execution plan for your model. It defines what routines you
        will execute and how the input and output will be handled.

        Parameters
        ----------
        data: dict or DataFrame
            the datatype is decided by the IO Specification; (df_to_df,
            dict_to_dict, etc.). 
        """
        pass

    def serve(self):
        """
        Creates a test server on port 5000 for testing purposes. This is a way to test
        your model before you deploy it to production.
        """
        template_folder = os.path.join(os.getcwd(), "templates")
        app = Flask(__name__, template_folder=template_folder)
        @app.route("/", methods=['GET', 'POST'])
        def testserver():
            if request.method=="GET":
                return render_template("demo.html")
            elif request.method=="POST":
                data = request.json
                try:
                    result = self.execute(data)
                    if isinstance(result, pd.DataFrame):
                        result = result.to_dict('list')
                    return jsonify(result)
                except Exception, e:
                    result = {"error": str(e)}
                    return jsonify(result)
            else:
                return "Not Implemented."
        app.run(debug=True)

    def run(self, testcase=None):
        """
        Runs a local 'Yhat Server' that mimics the production environment. Data
        can be passed to the server via stdin (by pasting data into the terminal
        ).
        
        If your data is too big to paste into the terminal, try setting up a 
        testcase which will run automatically. To do this, use the json module
        and run json.dumps({ YOUR DATA}) to generate acceptable input for Yhat.

        Parameters
        ----------
        testcase: optional, JSON string
            If a testcase is included, it will automatically execute when `run`
            is called.
        """
        print "Paste your test data here"
        print "Data should be formatted as valid JSON."
        print Fore.CYAN + "Hit <ENTER> to execute your model"
        print Fore.RED + "Press <CTRL + C> or type 'quit' to exit"
        print Fore.RESET + "="*40
        if testcase:
            sys.stdout.write("[In] %s\n" % testcase)
            data = parse_json(testcase + '\n')
            sys.stdout.write("[Out]\n")
            print self.execute(data)
        while True:
            sys.stdout.write("[In] ")
            data = sys.stdin.readline()
            if data.strip()=="quit":
                sys.exit()
            data = parse_json(data)
            sys.stdout.write("[Out]\n")
            print self.execute(data)


class Model(object):
    def __init__(self, **kwargs):
        for kw, arg in kwargs.iteritems():
            if kw=="requirements":
                arg = open(arg).read().strip()
            setattr(self, kw, arg)

class BaseModel(Model):
    __name__ = "BaseModel"
    """
    BaseModel is the standard Yhat model class that gives your model
    the functionality to run it as a RESTful API. Once you create your own
    model, you need to implement the following functions:
         - transform
         - predict
    """
    def require(self):
        """
        Define what libraries and modules you want to include in your project
        ====================================================================
        def require(self):
            from StringIO import StringIO
            from string import letters
        """
        pass

    def transform(self, rawData):
        """
        Transform takes the raw data that's going to be sent to your yhat API
        and converts it into the format required to be run through your model.
        """
        pass

    def predict(self, transformedData):
        """
        Predict executes your predictive model, formats the data into response,
        and returns it.
        """
        pass

    def execute(self, data):
        self.require()
        data = self.transform(data)
        return self.predict(data)


class StepModel(Model):
    __name__ = "StepModel"
    """
    StepModel allows you to define steps as methods. Each step is numbered
    (i.e. step_1, step_2, step_3, etc.). Steps are executed in order with the 
    output from the previous step becoming the input to the next step.

    Parameters
    ----------
    
    Examples
    --------

    class MyModel(StepModel):
        def step_1(self, data):
            return 1
        def step_2(self, data):
            return [data, 2]
        def step_3(self, data):
            return data + [3]
    mm = MyModel()
    print mm.execute(None)
    # [1, 2, 3]
    # This executes like so...
    # one = mm.step_1(None)
    # two = mm.step_2(one)
    # three = mm.step_3(two)
    # print three
    """
    def require(self):
        pass

    def _get_steps(self):
        steps = []
        for name, step in inspect.getmembers(self, predicate=inspect.ismethod):
            if re.match("step_", name):
                order = name.lstrip("step_")
                try:
                    order = int(order)
                except:
                    raise Exception("'%s' - step names must be integers" % name)
                steps.append((order, name, step))
        steps = sorted(steps, key=lambda x: int(x[0]))
        return steps
    
    def execute(self, data):
        for _, name, step in self._get_steps():
            try:
                data = step(data)
            except Exception, e:
                print "Error executing step: '%s'" % name
                print "Exception: %s" % str(e)
                return
        return data




import pip
import shutil
import urllib2
import json
import os

from utils import download_file
import credentials

PROJECT_DIR = os.path.join(os.environ['HOME'], ".yhat/templates")

def setup(template_name, project_name):
    """
    Return a DataFrame with the same shape as self and whose corresponding
    entries are from self where cond is True and otherwise are from other.

    Parameters
    ----------
    template_name: string
        name of the template. this is the name of a pre-designed template
    project_name: string
        name of the project. this will be the name of the directory created on 
        your local machine
    """
    if template_name=='*new project*':
        return setup_new_project()
    elif template_name=="*download project*":
        return find_template(project_name)
    replacements = {
        "project_name": project_name
    }
    replacements.update(credentials.read())
    config_filepath = os.path.join(PROJECT_DIR, template_name + '.json')
    config = json.load(open(config_filepath))
    if os.path.isdir(project_name):
        shutil.rmtree(project_name)
    print "creating project %s" % project_name
    os.mkdir(project_name)
    for dirname in config.get("dirs", []):
        path = os.path.join(project_name, dirname)
        if os.path.isdir(path)==False:
            os.mkdir(path)
    for newfile in config.get("files", []):
        filename = os.path.join(project_name, newfile['name'])
        with open(filename, "w") as f:
            f.write(newfile['content'].format(**replacements))

    print "Downloading data dependencies for project"
    print "="*80
    for datafile in config.get("data", []):
        print " >>> %s" % datafile['name']
        download_file(datafile['source'], datafile['name'])

    reqs_filepath = os.path.join(project_name, "reqs.txt")
    try:
        reqs = open(reqs_filepath).read().split('\n')
        pip_args = ['-vvv']
        pip_args.append('install')
        for req in reqs:
            pip_args.append(req)
        pip.main(initial_args=pip_args)
    except:
        print "%s not found!" % reqs_filepath

def setup_new_project():
    """
    # NOT IMPLEMENTED
    """
    inputs = [
        {"prompt": "Project Name: ", "variable": "name"},
        {"prompt": "Description: ", "variable": "description"},
        {"prompt": "GitHub username: ", "variable": "github_username"},
        {"prompt": "Your name: ", "variable": "your_name"},
        {"prompt": "Your email: ", "variable": "your_email"}
    ]
    user_input = {}
    for prompt in inputs:
        user_input[prompt['variable']] = raw_input(Fore.CYAN + prompt['prompt'])
    
    return user_input

def find_template(query):
    """
    Looks through the index of Yhat templates. If any exist, they are yielded
    in a generator.

    Parameters
    ----------
    query: string
        search parameter

    Returns
    -------
    templates: generator
        any matching template(s)
    """
    ALL_TEMPLATES = [
        {"name": "something"}
    ]
    for template in ALL_TEMPLATES:
        if re.find(query, json.dumps(template)):
            yield template

def download_template(template_source):
    """
    Downloads a template and saves it to the .yhat/templates directory

    template_source: string
        URL of the .json template file
    """
    if template_source.endswith(".json"):
        filename = os.path.basename(template_source)
        filename = os.path.join(os.environ['HOME'], '.yhat/templates', filename)
        download_file(template_source, filename)

def bundle(name, directory):
    """
    Grab each file and put it in a list with it's pathname (relative to 
    directory) and it's contents. Create a list of each directory (and any
    sub-directories). Create a list of data elements. These should have a 
    name and a source (url for now).


    Parameters
    ----------
    name: string
        the name of the project being bundled
    directory: string
        location of the project
    """
    
    project_file = os.path.join(directory, name + ".json")
    project = {
        "name": name,
        "dirs": [],
        "files": [],
        "data": []
    }
    for i, (dirname, subdirs, files) in enumerate(os.walk(directory)):
        subdir = dirname.replace(directory, '').strip('/')
        if subdir:
            project["dirs"].append(subdir)
        for f in files:
            # TODO: make this more comprehensive; maybe a .gitignore?
            if f.endswith(".pyc"):
                continue
            f = os.path.join(dirname, f)
            # we don't want to bundle the bundle
            if os.path.basename(f)==os.path.basename(project_file):
                continue
            elif dirname==os.path.join(directory, "data"):
                # TODO: add option to upload to AWS/Dropbox
                source = raw_input("Input source URL for %s: " % f)
                f = f.replace(directory, '')
                f = f.strip('/')
                project["data"].append({"name": f, "source": source})
            else:
                content = open(f).read()
                f = f.replace(directory, '')
                f = f.strip('/')
                project["files"].append({"name": f, "content": content})
    with open(project_file, "wb") as f:
        json.dump(project, f, indent=2) 




import sys
import warnings
import base64
import json
import pickle
import urllib2, urllib
import types
import websocket
import uuid
import zlib
import re
import os
from colorama import init
from colorama import Fore

init()

from deployment.models import YhatModel
from deployment.save_session import save_function, _get_source

try:
    import pandas as pd
except:
    warnings.warn("Could not import pandas")

BASE_URI = "http://api.yhathq.com/"


class API(object):
    """
    An abstract class that implements some of the more generic methods for 
    interacting with REST APIs.

    Parameters
    ----------
    base_uri: string
        URI of the API that you're working with
    """
    def __init__(self, base_uri):
        self.base_uri = base_uri
        self.headers = {'Content-Type': 'application/json'}

    def get(self, endpoint, params):
        """
        Parameters
        ----------
        endpoint: string
            name of the rest endpoint you want to hit
        params: dictionary
            querystring parameters for API call

        Returns
        -------
        data: dictionary
            whatever is returned by the API
        """
        try:
            url = self.base_uri + endpoint + "?" + urllib.urlencode(params)
            req = urllib2.Request(url)
            req.add_header('Content-Type', 'application/json')
            auth = '%s:%s' % (params['username'], params['apikey'])
            base64string = base64.encodestring(auth).replace('\n', '')
            req.add_header("Authorization", "Basic %s" % base64string)
            response = urllib2.urlopen(req)
            rsp = response.read()
            return json.loads(rsp)
        except Exception, e:
            raise e
    
    def post(self, endpoint, params, data):
        """
        Parameters
        ----------
        endpoint: string
            name of the rest endpoint you want to hit
        params: dictionary
            querystring parameters for API call
        data: dictionary
            data you want transfered as JSON

        Returns
        -------
        data: dictionary
            whatever is returned by the API
        """
        try:
            url = self.base_uri + endpoint + "?" + urllib.urlencode(params)
            req = urllib2.Request(url)
            req.add_header('Content-Type', 'application/json')
            auth = '%s:%s' % (params['username'], params['apikey'])
            base64string = base64.encodestring(auth).replace('\n', '')
            req.add_header("Authorization", "Basic %s" % base64string)
            try:
                data = json.dumps(data)
            except Exception, e:
                msg = """Whoops. The data you're trying to send could not be 
converted into JSON. If the data you're attempting to send includes a numpy 
array, try casting it to a list (x.tolist()). If you're still having trouble, 
read our help here: {URL}.""".format(URL="http://docs.yhathq.com/help/json")
                print msg
            response = urllib2.urlopen(req, data)
            rsp = response.read()
            return json.loads(rsp)
        except Exception, e:
            raise e
    
    def handshake(self, model_name):
        """
        Parameters
        ----------
        model_name: string
            name of the model you want to connect to

        Returns
        -------
        ws: WebSocket connection
            a connection to the model's WebSocketServer
        """
        ws_uri = "{BASE_URI}/{USERNAME}/models/{MODEL_NAME}/"
        ws_base = self.base_uri.replace("http://", "ws://")
        ws_uri = ws_uri.format(BASE_URI=ws_base, USERNAME=self.username,
                               MODEL_NAME=model_name)
        ws = websocket.create_connection(ws_uri)
        auth = { 
                "username": self.username, 
                "apikey": self.apikey
        }
        ws.send(json.dumps(auth))
        return ws
 

class Yhat(API):
    """
    A connection to the Yhat API

    Parameters
    ----------
    username: string
        your Yhat username
    apikey: string
        your Yhat apikey
    uri: string
        your Yhat server URL (i.e. http://cloud.yhathq.com/)
    """
    def __init__(self, username, apikey, uri):
        self.username = username
        self.apikey = apikey
        if uri.endswith("/")==False:
            uri += "/"
        self.ws = None
        self.base_uri = uri
        self.headers = {'Content-Type': 'application/json'}
        self.q = {"username": self.username, "apikey": apikey}
        if self.base_uri!=BASE_URI:
            if self._authenticate()==False:
                raise Exception("Incorrect username/apikey!")

    def _check_obj_size(self, obj):
        """
        Checks if an object is greater than 50 MB

        Parameters
        ----------
        obj: object
            any object

        Returns
        -------
        boolean
        """
        if self.base_uri!=BASE_URI:
            # not deploying to the cloud so models can be as big as you want
            if sys.getsizeof(obj) > 52428800:
                return False
        elif sys.getsizeof(obj) > 52428800:
            raise Exception("Sorry, your file is too big for a free account.")

        return True
    
    def _authenticate(self):
        """
        Returns
        -------
        authed: boolean
            verifies your API credentials are valid
        """
        authed = True
        try: 
            response = self.post('verify', self.q, {})
            error = response["success"];
        except Exception, e:
            authed = False
        return authed

    def _convert_to_json(self, data):
        """
        Tries to convert data into a JSON frieldy dict.

        Parameters
        ----------
        data: object
            data to be converted to valid JSON

        Returns
        -------
        data: object
            same data, but compatible with JSON
        """
        if isinstance(data, pd.DataFrame):
            data = data.to_dict('list')
        return data

    def predict(self, model, data):
        """
        Executes a single prediction via the Yhat API.

        Parameters
        ----------
        model: string
            the name of your model
        data: dictionary/data frame
            data required to make a single prediction. this can be a dict or
            a dataframe

        Returns
        -------
        data: dictionary
            data returned by a model
        """
        data = self._convert_to_json(data)
        q = self.q
        q['model'] = model
        if self.base_uri!=BASE_URI:
            endpoint = "%s/models/%s/" % (self.username, model)
        else:
            data = {"data": data}
            endpoint = 'predict'
        return self.post(endpoint, q, data)
    
    def predict_ws(self, data):
        """

        Parameters
        ----------
        data: dictionary or data frame
            data required to make a single prediction. this can be a dict or
            a dataframe

        Returns
        -------
        id: string
            the id of the event; this will come in handy when it
            is receieved by the WebSocket client (remember, this is 
            asynchronous)
        """
        if self.ws==None:
            msg = """In order to make predictions with WebScokets, you
need to connect to the server first. try running "connect_to_socket"
"""
            raise Exception(msg)
        data = self._convert_to_json(data)
        data['_id'] = str(uuid.uuid4())
        self.ws.send(json.dumps(data))
        return data['_id']

    def yield_results(self):
        """
        Listens to the WebSocket and awaits completed predictions

        Returns
        -------
        iterator: generator
            yields a new prediction
        """
        if self.ws is None:
            msg = """In order to make predictions with WebScokets, you
need to connect to the server first. try running "connect_to_socket"
"""
            raise Exception(msg)

        while True:
            yield self.ws.recv()

    def connect_to_socket(self, model):
        """
        Connects to the model's WebSocket endpoint. This is a pre-requisite for
        making predictions via the WebSocketServer.

        model: string
            name of the model you want to connect to
        """
        self.ws = self.handshake(model)

    def _extract_model(self, name, model, session):
        """
        Extracts source code and any objects required to deploy the model.

        Parameters
        ----------
        name: string
            name of your model
        model: YhatModel
            an instance of a Yhat model
        session: globals()
            your Python's session variables (i.e. "globals()")
        """
        if 1==2 and _get_source(YhatModel.execute)==_get_source(model.execute):
            msg = "'execute' method was not implemented. If you believe that you did "
            msg += "implement the 'execute' method, check to make sure there isn't an "
            msg += "indentation error."
            raise Exception(msg)

        
        bundle = save_function(model, session)
        bundle["largefile"] = True
        bundle["username"] = self.username
        bundle["language"] = "python"
        bundle["modelname"] = name
        bundle["className"] = model.__name__
        bundle["reqs"] = getattr(model, "REQUIREMENTS", "")
        # make sure we freeze Yhat so we're sure we're using the right version
        # this makes it a lot easier to upgrade the client
        import yhat
        bundle["reqs"] += '\n' + "yhat==" + yhat.__version__
        bundle["reqs"] = bundle["reqs"].strip()
        return bundle
    
    def deploy(self, name, model, session, sure=False):
        """
        Deploys your model to a Yhat server

        Parameters
        ----------
        name: string
            name of your model
        model: YhatModel
            an instance of a Yhat model
        session: globals()
            your Python's session variables (i.e. "globals()")
        sure: boolean
            if true, then this will force a deployment (like -y in apt-get).
            if false or blank, this will ask you if you're sure you want to
            deploy
        """
        # first let's check and make sure the user actually wants to deploy
        # a new version
        if sure==False:
            sure = raw_input("Are you sure you want to deploy? (y/N): ")
            if sure.lower()!="y":
                print "Deployment canceled"
                sys.exit()
        bundle = self._extract_model(name, model, session)
        if self._check_obj_size(bundle)==False:
            # we're not going to deploy; model is too big, but let's give the 
            # user the option to upload it manually
            print "Model is to large to deploy over HTTP"
            should_we_deploy = raw_input("Would you like to upload manually? (Y/n): ")
            if should_we_deploy.lower()=="y" or should_we_deploy=="":
                self.deploy_to_file(name, model, session)
        else:
            # upload the model to the server
            # TODO: upload w/ a progress bar
            return self.post("deployer/model", self.q, bundle)

    def deploy_to_file(self, name, model, session, compress=True):
        """
        Bundles a local version of your model that can be manually uploaded to
        the server.

        Parameters
        ----------
        name: string
            name of your model
        model: YhatModel
            an instance of a Yhat model
        session: globals()
            your Python's session variables (i.e. "globals()")
        """
        bundle = self._extract_model(name, model, session)
        with open("%s.yhat" % name, "w") as f:
            bundle = json.dumps(bundle)
            if compress==True:
                bundle = zlib.compress(bundle)
            f.write(bundle)

        print "Model successfully bundled to file:"
        print "\t%s/%s.yhat" % (os.getcwd(), name)
        msg = "To deploy, visit %s and upload %s."
        upload_url = os.path.join(self.base_uri, "model", "upload")
        msg = msg % (upload_url, "%s.yhat" % name)
        print Fore.CYAN + msg
        print Fore.RESET

if __name__=="__main__":
    import pandas as pd
    df = pd.DataFrame({
        "x": range(10),
        "y": range(10)
    })
    yh = Yhat("blah", "blue")
    print yh._convert_to_json({"x": 1})
    print yh._convert_to_json(df)

    from yhat import credentials
    creds = credentials.read()

    yh = Yhat(creds['username'], creds['apikey'], "http://localhost:8080/")

    print yh.predict("mymodel", {"beer": "Coors Light"})
    yh.connect_to_socket("mymodel")
    print yh.predict_ws({"beer": "Coors Light"})
    for item in yh.yield_results():
        print item




import urllib2
import os

from progressbar import ProgressBar, Bar, ETA, Percentage


def download_file(url, filename):
    """
    Downloads a file and displays a progress bar on the screen.

    url: string
        URL of the file's source
    filename: string
        location where the downloaded file should be saved
    """
    file_conn = urllib2.urlopen(url)
    filesize = file_conn.headers['content-length']
    processed = 0
    pbar = ProgressBar(widgets=[Bar(), ' ', ETA(), ' ', Percentage()],
            maxval=int(filesize)).start()
    with open(filename, "wb") as f:
        for line in file_conn:
            processed += len(line)
            pbar.update(processed)
            f.write(line)
    print
from progressbar import ProgressBar, Bar, ETA, Percentage


class file_with_callback(file):
    def __init__(self, path, mode, *args):
        file.__init__(self, path, mode)
        self.seek(0, os.SEEK_END)
        self._total = self.tell()
        self.seek(0)
        self.pbar = ProgressBar(widgets=[Bar(), ' ', ETA(), ' ', Percentage()],
                maxval=self._total).start()
        self.processed = 0
        self._callback = self.pbar.update

    def __len__(self):
        return self._total

    def read(self, size):
        data = file.read(self, size)
        self.processed += len(data)
        self._callback(self.processed)
        return data

def upload_file(url, filename):
    stream = file_with_callback(filename, 'rb', filename)
    req = urllib2.Request(url, stream)
    res = urllib2.urlopen(req)
    print




import json
import base64
import os


def setup():
    """
    Prompts the user for their credentials and the saves them to a Yhat "dot"
    file.
    """
    username = raw_input("Yhat username: ")
    apikey = raw_input("Yhat apikey: ")
    with open(os.path.join(os.environ['HOME'], '.yhat', '.config'), 'w') as f:
		data = json.dumps({"username": username, "apikey": apikey})
		data = base64.encodestring(data)
		f.write(data)

def read():
    """
    Extracts credentials from a "dot" file
    
    Returns
    =======
    credentials: dict
        your credentials in form:
        {"username": "YOUR USERNAME", "apikey": "YOUR APKIKEY}"
    """
    data = open(os.path.join(os.environ['HOME'], '.yhat', '.config')).read()
    return json.loads(base64.decodestring(data))

